import numpy as np
import pytest
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from pyfm2d import WaveTrackerOptions, display_model, BasisModel, calc_wavefronts
from pyfm2d.wavetracker import (
    _calc_wavefronts_process,
    _calc_wavefronts_multithreading,
    cleanup,
    _build_velocity_grid,
)
import pyfm2d.fastmarching as fmm

PLOT = False
HOMOGENOUS_VELOCITY = 2.0


def get_sources():
    # to ensure separation of sources and receivers
    # get some random sources in one quadrant
    return np.random.uniform(0.05, 0.45, (5, 2))


def get_receivers():
    # to ensure separation of sources and receivers
    # get some random receivers in the opposite quadrant
    return np.random.uniform(0.55, 0.95, (3, 2))


def create_velocity_grid_model():
    # Create a simple velocity model that we can easily
    # manually calculate the travel times for.
    # In this case, we have a 5x5 grid with a velocity of 2.0
    # so the travel time from the source to the receiver should be
    # 0.5 times the path length.
    m = np.ones((5, 5)) * HOMOGENOUS_VELOCITY
    g = BasisModel(m)
    return g


def calculate_expected_tt(src, rec):
    diff = (src[:, np.newaxis] - rec).reshape(-1, 2)  # some broadcasting magic
    return np.sqrt(np.sum(diff**2, axis=1)) / HOMOGENOUS_VELOCITY

def convert_kms_2_deg(src,rec,extent):
    degperkms = 180./(6371.0*np.pi)
    rec = degperkms*rec
    src = degperkms*src
    ext = [i*degperkms for i in extent] # unit box in kms converted to degrees
    return src,rec,ext

def test__calc_wavefonts_process():
    g = create_velocity_grid_model()
    recs = get_receivers()
    srcs = get_sources()
    cartesian = False
    extent = [0., 1., 0., 1.]

    # Check the travel times
    expected_tt = calculate_expected_tt(srcs, recs)

    #srcs,recs,extent = convert_kms_2_deg(srcs,recs,extent) # For Spherical model we should convert the input spatial units to degrees

    options = WaveTrackerOptions(times=True, paths=True, frechet=True, cartesian=True)
    result = _calc_wavefronts_process(
        g.get_velocity(),
        recs,
        srcs,
        extent=extent,
        options=options,
    )

    assert result.ttimes is not None
    assert result.paths is not None
    assert result.frechet is not None

    # fmm seems quite inaccurate because of the small grid size
    # and putting float32 everywhere
    print('expected_tt',expected_tt)
    print('result.ttimes',result.ttimes)
    assert np.allclose(result.ttimes, expected_tt, atol=1e-2)

    if PLOT:
        display_model(g.get_velocity(), paths=result.paths)


def test_cleanup():

    @cleanup
    def failing_function():
        # setup required to allocate arrays in the first place
        v = create_velocity_grid_model().get_velocity()
        recs = get_receivers()
        srcs = get_sources()
        options = WaveTrackerOptions(times=True, paths=True, frechet=True)
        fmm.set_solver_options(
            options.dicex,
            options.dicey,
            options.sourcegridrefine,
            options.sourcedicelevel,
            options.sourcegridsize,
            options.earthradius,
            options.schemeorder,
            options.nbsize,
            options.lttimes,
            options.lfrechet,
            options.tsource,
            options.lpaths,
            options.lcartesian,
            int(options.quiet),
        )

        fmm.set_sources(srcs[:, 1], srcs[:, 0])  # ordering inherited from fm2dss.f90
        fmm.set_receivers(recs[:, 1], recs[:, 0])  # ordering inherited from fm2dss.f90

        nvx, nvy = v.shape
        # grid node spacing in lat and long
        extent = [0, 1, 0, 1]
        dlat = (extent[3] - extent[2]) / (nvy - 1)
        dlong = (extent[1] - extent[0]) / (nvx - 1)

        vc = _build_velocity_grid(v)

        fmm.set_velocity_model(nvy, nvx, extent[3], extent[0], dlat, dlong, vc)

        # set up time calculation between all sources and receivers
        associations = np.ones((recs.shape[0], srcs.shape[0]))
        fmm.set_source_receiver_associations(associations)

        fmm.allocate_result_arrays()
        raise ValueError("This function should fail")

    with pytest.raises(ValueError):
        failing_function()

    # If the clean up worked, we should now be able to
    # allocate again without any issues
    fmm.allocate_result_arrays()
    # Now cleanup again to avoid any issues with the next test
    fmm.deallocate_result_arrays()


def test_calc_wavefonts_multithreading():
    g = create_velocity_grid_model()
    recs = get_receivers()
    srcs = np.concatenate([get_sources() for _ in range(4)])
    extent = [0.,1.,0.,1.]

    # Check the travel times
    expected_tt = calculate_expected_tt(srcs, recs)

    srcs,recs,extent = convert_kms_2_deg(srcs,recs,extent) # For Spherical model we should convert the input spatial units to degrees

    options = WaveTrackerOptions(times=True, paths=True, frechet=True)
    result = _calc_wavefronts_multithreading(
        g.get_velocity(),
        recs,
        srcs,
        extent=extent,
        options=options,
        nthreads=4,
    )

    assert result.ttimes is not None
    assert result.paths is not None
    assert result.frechet is not None

    assert np.allclose(result.ttimes, expected_tt, atol=1e-2)

    if PLOT:
        display_model(g.get_velocity(), paths=result.paths)


def test_user_provided_threadpool_executor():
    """Test that ThreadPoolExecutor raises an error due to Fortran shared memory conflicts"""
    g = create_velocity_grid_model()
    recs = get_receivers()
    srcs = get_sources()
    extent = [0., 1., 0., 1.]
    
    srcs, recs, extent = convert_kms_2_deg(srcs, recs, extent)
    
    options = WaveTrackerOptions(times=True, paths=True, frechet=True)
    
    # ThreadPoolExecutor cannot be used due to shared memory conflicts in the Fortran implementation
    # Multiple threads trying to allocate the same global Fortran arrays causes segfaults
    with pytest.raises(ValueError, match="ThreadPoolExecutor is not supported"):
        with ThreadPoolExecutor(max_workers=2) as pool:
            result = calc_wavefronts(
                g.get_velocity(),
                recs,
                srcs,
                pool=pool,
                extent=extent,
                options=options,
            )


def test_user_provided_processpool_executor():
    """Test calc_wavefronts with user-provided ProcessPoolExecutor"""
    g = create_velocity_grid_model()
    recs = get_receivers()
    srcs = get_sources()
    extent = [0., 1., 0., 1.]
    
    # Calculate expected travel times before conversion
    expected_tt = calculate_expected_tt(srcs, recs)
    
    srcs, recs, extent = convert_kms_2_deg(srcs, recs, extent)
    
    options = WaveTrackerOptions(times=True, paths=True, frechet=True)
    
    with ProcessPoolExecutor(max_workers=2) as pool:
        result = calc_wavefronts(
            g.get_velocity(),
            recs,
            srcs,
            pool=pool,
            extent=extent,
            options=options,
        )
    
    assert result.ttimes is not None
    assert result.paths is not None
    assert result.frechet is not None
    
    assert np.allclose(result.ttimes, expected_tt, atol=1e-2)


def test_pool_reuse():
    """Test that a user-provided pool can be reused for multiple calculations"""
    g = create_velocity_grid_model()
    recs = get_receivers()
    srcs1 = get_sources()[:2]
    srcs2 = get_sources()[2:4]
    extent = [0., 1., 0., 1.]
    
    srcs1, _, extent1 = convert_kms_2_deg(srcs1, recs, extent)
    srcs2, recs, extent2 = convert_kms_2_deg(srcs2, recs, extent)
    
    options = WaveTrackerOptions(times=True, paths=False, frechet=False)
    
    with ProcessPoolExecutor(max_workers=2) as pool:
        # First calculation
        result1 = calc_wavefronts(
            g.get_velocity(),
            recs,
            srcs1,
            pool=pool,
            extent=extent1,
            options=options,
        )
        
        # Second calculation with same pool
        result2 = calc_wavefronts(
            g.get_velocity(),
            recs,
            srcs2,
            pool=pool,
            extent=extent2,
            options=options,
        )
    
    assert result1.ttimes is not None
    assert result2.ttimes is not None
    assert len(result1.ttimes) == len(srcs1) * len(recs)
    assert len(result2.ttimes) == len(srcs2) * len(recs)


def test_calc_wavefonts_multithreading_vs_serial():
    g = create_velocity_grid_model()
    recs = get_receivers()
    srcs = get_sources()
    extent = [0.,1.,0.,1.]

    srcs,recs,extent = convert_kms_2_deg(srcs,recs,extent) # For Spherical model we should convert the input spatial units to degrees

    options = WaveTrackerOptions(times=True, paths=True, frechet=True)
    result_serial = _calc_wavefronts_process(
        g.get_velocity(),
        recs,
        srcs,
        extent=extent,
        options=options,
    )

    result_parallel = _calc_wavefronts_multithreading(
        g.get_velocity(),
        recs,
        srcs,
        extent=extent,
        options=options,
        nthreads=4,
    )

    assert np.allclose(result_serial.ttimes, result_parallel.ttimes)
    for i in range(len(result_serial.paths)):
        assert np.allclose(result_serial.paths[i], result_parallel.paths[i]), f"Path {i} is not equal"
    assert np.allclose(result_serial.frechet.toarray(), result_parallel.frechet.toarray())



import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from scipy.interpolate import RectBivariateSpline
from scipy.sparse import csr_matrix, vstack
import faulthandler
from dataclasses import dataclass
from typing import Any, Callable, Optional
import concurrent.futures
from functools import reduce
import operator
import sys

from . import fastmarching as fmm
from . import bases as base


faulthandler.enable()

# --------------------------------------------------------------------------------------------

# This library is a python interface to Nick Rawlinson's 2D Fast Marching Fortran package fm2dss.f90
#
# It uses the ctypes interface to fm2dss.f90 developed by Juerg Hauser in file "._pyfm2dss"
#
# History:
#        January 2025:  Uses ctypes interface in pyfm2dss.py
#        January 2024:  Updated to interface with package bases.py allowing multiple 2D model bases
#                       including pixel and discrete cosine. Also interfaces with Jacobian integration
#                       package jip.py to calculate Frechect kernels when model bases
#                       are not simple pixel bases, e.g discrete cosine bases.
#
#        Definitions within waveTracker follow conventions from Andrew Valentine's rayTracer.py package.
#
# M. Sambridge
# January 2025
# --------------------------------------------------------------------------------------------


class Error(Exception):
    """Base class for other exceptions"""

    pass


class InputError(Exception):
    """Raised when necessary inputs are missing"""

    def __init__(self, msg=""):
        super().__init__(msg)


@dataclass
class WaveTrackerResult:
    """
    Result class for the calc_wavefronts function.

    Attributes:

        ttimes (np.ndarray): first arrival travel times between ns sources and nr receivers.

        paths (list): list of 2-D arrays (x,y) for ray paths between ns sources and nr receivers.

        ttfield (np.ndarray): 2-D array of travel time field for source tfieldsource at resolution mx*my

        frechet (csr_matrix): 2D array of shape (nrays, nx*ny) in sparse csr format containing derivatives of travel

    """

    ttimes: Optional[np.ndarray] = None
    paths: Optional[list] = None
    ttfield: Optional[np.ndarray] = None
    frechet: Optional[csr_matrix] = None

    def __add__(self, other: "WaveTrackerResult"):

        try:
            self._check_compatibility(other)
        except InputError as e:
            raise InputError(f"Incompatible WaveTrackerResults: {e}")

        if self.ttimes is not None:
            ttimes = np.concatenate([self.ttimes, other.ttimes])
        else:
            ttimes = None
        if self.paths is not None:
            paths = self.paths + other.paths
        else:
            paths = None
        if self.ttfield is not None:
            ttfield = np.concatenate([self.ttfield, other.ttfield])
        else:
            ttfield = None
        if self.frechet is not None:
            frechet = vstack([self.frechet, other.frechet])
        else:
            frechet = None

        return WaveTrackerResult(ttimes, paths, ttfield, frechet)

    def _check_compatibility(self, other: "WaveTrackerResult"):
        if (self.ttimes is None) != (other.ttimes is None):
            raise InputError("Travel times are not available for both results.")
        if (self.paths is None) != (other.paths is None):
            raise InputError("Ray paths are not available for both results.")
        if (self.ttfield is None) != (other.ttfield is None):
            raise InputError("Travel time fields are not available for both results.")
        if (self.frechet is None) != (other.frechet is None):
            raise InputError("Frechet derivatives are not available for both results.")


@dataclass
class WaveTrackerOptions:
    """
    WaveTrackerOptions is a configuration class for the calc_wavefronts function.

    Attributes:

        times (bool): Whether to calculate travel times. Default is True.

        paths (bool): Whether to calculate ray paths. Default is False.

        frechet (bool): Whether to compute Frechet derivatives. Default is False.

        ttfield_source (int): Source index for computation of travel time field. If <0 then no fields are computed. Default is -1.

        sourcegridrefine (bool): Apply sourcegrid refinement. Default is True.

        sourcedicelevel (int): Source discretization level. Number of sub-divisions per cell (default=5, i.e. 1 model cell becomes 5x5 sub-cells)

        sourcegridsize (int): Number of model cells to refine about source at sourcedicelevel (default=10, i.e. 10x10 cells are refined about source)

        earthradius (float): radius of Earth in km, used for spherical to Cartesian transform. Default is 6371.0.

        schemeorder (int): switch to use first order (0) or mixed order (1) scheme. Default is 1.

        nbsize (float): Narrow band size (0-1) as fraction of nnx*nnz. Default is 0.5.

        cartesian (bool): True if using a Cartesian spatial frame. Default is False.

        velocityderiv (bool): Switch to return Frechet derivatives of travel times w.r.t. velocities (True) rather than slownesses (False). Default is False.

        dicex (int): x-subgrid discretization level for B-spline interpolation of input mode. Default is 8.

        dicey (int): y-subgrid discretization level for B-spline interpolation of input model. Default is 8.

        quiet (bool): Suppress non-fatal ray path and boundary warnings. Default is False (show warnings).
    """

    times: bool = True
    paths: bool = False
    frechet: bool = False
    ttfield_source: int = -1
    sourcegridrefine: bool = True
    sourcedicelevel: int = 5
    sourcegridsize: int = 10
    earthradius: float = 6371.0
    schemeorder: int = 1
    nbsize: float = 0.5
    cartesian: bool = False
    velocityderiv: bool = False
    dicex: int = 8
    dicey: int = 8
    quiet: bool = False

    def __post_init__(self):
        # mostly convert boolean to int for Fortran compatibility

        # Write out ray paths. Only allow all (-1) or none (0)
        self.lpaths: int = -1 if self.paths else 0

        # int to calculate travel times (y=1,n=0)
        self.lttimes: int = 1 if self.times else 0

        # int to calculate Frechet derivatives of travel times w.r.t. slownesses (0=no,1=yes)
        self.lfrechet: int = 1 if self.frechet else 0

        # int to calculate travel fields (0=no,1=all)
        self.tsource: int = 1 if self.ttfield_source >= 0 else 0

        # int to activate cartesian mode (y=1,n=0)
        self.lcartesian: int = 1 if self.cartesian else 0

def cleanup(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            fmm.deallocate_result_arrays()
            raise e

    return wrapper


@cleanup
def calc_wavefronts(
    v,
    recs,
    srcs,
    extent=[0.0, 1.0, 0.0, 1.0],
    options: Optional[WaveTrackerOptions] = None,
    nthreads: int = 1,
    pool=None,
    quiet: Optional[bool] = None,
):
    """

    A function to perform 2D Fast Marching of wavefronts from sources in a 2D velocity model.

    Inputs:
        v, ndarray(nx,ny)          : coefficients of velocity field in 2D grid with dimension (nx,ny).
        recs, ndarray(nr,2)        : receiver locations (x,y). Where nr is the number of receivers.
        srcs, ndarray(ns,2)        : source locations (x,y). Where ns is the number of receivers.
        extent, list               : 4-tuple of model extent [xmin,xmax,ymin,ymax]. (default=[0.,1.,0.,1.])
        options, WaveTrackerOptions: configuration options for the wavefront tracker. (default=None)
        nthreads, int              : number of threads to use for multithreading. Multithreading is performed over sources (default=1)
        pool                       : User-provided pool for parallel processing. If provided, this takes precedence
                                    over the nthreads parameter. The pool must implement either a submit() method
                                    (like concurrent.futures executors) or a map() method (like schwimmbad pools).
                                    When providing a pool, the user is responsible for its lifecycle management.
                                    (default=None)
        quiet, bool                : Suppress non-fatal ray path and boundary warnings. If provided, overrides
                                    options.quiet. (default=None)


    Returns
        WaveTrackerResult: a dataclass containing the results of the wavefront tracking.

    Notes:
        Internally variables are converted to np.float32 to be consistent with Fortran code fm2dss.f90.

    """

    # Initialize options if not provided
    if options is None:
        options = WaveTrackerOptions()

    # Direct parameter overrides options.quiet if provided
    if quiet is not None:
        import dataclasses
        options = dataclasses.replace(options, quiet=quiet)

    if pool is not None:
        # Check if pool is a ThreadPoolExecutor
        from concurrent.futures import ThreadPoolExecutor
        if isinstance(pool, ThreadPoolExecutor):
            raise ValueError(
                "ThreadPoolExecutor is not supported due to shared memory conflicts in the "
                "underlying Fortran implementation. Multiple threads cannot safely allocate "
                "the same global Fortran arrays. Please use ProcessPoolExecutor or other "
                "process-based pools instead."
            )
        # User-provided pool takes precedence
        return _calc_wavefronts_multithreading(v, recs, srcs, extent, options, pool=pool)
    elif nthreads <= 1:
        return _calc_wavefronts_process(v, recs, srcs, extent, options)
    else:
        return _calc_wavefronts_multithreading(v, recs, srcs, extent, options, nthreads=nthreads)


def _calc_wavefronts_process(
    v,
    recs,
    srcs,
    extent=[0.0, 1.0, 0.0, 1.0],
    options: Optional[WaveTrackerOptions] = None,
):
    # here extent[3],extent[2] is N-S range of grid nodes
    #      extent[0],extent[1] is W-E range of grid nodes
    if options is None:
        options = WaveTrackerOptions()

    recs = recs.reshape(-1, 2)  # ensure receiver array is 2D and float32
    srcs = srcs.reshape(-1, 2)  # ensure source array is 2D and float32

    _check_sources_receivers_inside_extent(srcs, recs, extent)

    _check_requested_source_exists(options.ttfield_source, len(srcs))

    fmm.set_solver_options(
        options.dicex,
        options.dicey,
        options.sourcegridrefine,
        options.sourcedicelevel,
        options.sourcegridsize,
        options.earthradius,
        options.schemeorder,
        options.nbsize,
        options.lttimes,
        options.lfrechet,
        options.tsource,
        options.lpaths,
        options.lcartesian,
        int(options.quiet),
    )

    fmm.set_sources(srcs[:, 1], srcs[:, 0])  # ordering inherited from fm2dss.f90
    fmm.set_receivers(recs[:, 1], recs[:, 0])  # ordering inherited from fm2dss.f90

    nvx, nvy = v.shape
    # grid node spacing in lat and long
    dlat = (extent[3] - extent[2]) / (nvy - 1)
    dlong = (extent[1] - extent[0]) / (nvx - 1)

    vc = _build_velocity_grid(v)

    if (options.lcartesian == 1): # grid in regular order if Cartesian mode (co-ords are in kms)
        fmm.set_velocity_model(nvy, nvx, extent[2], extent[0], dlat, dlong, vc)

    else:                         # y-grid (Lat) required in reversed order if Spherical mode (co-ords are in degs)

        vc = vc[:, ::-1] # reverse direction of velocity model in latitude direction for Spherical model
        fmm.set_velocity_model(nvy, nvx, extent[3], extent[0], dlat, dlong, vc)

    # set up time calculation between all sources and receivers
    associations = np.ones((recs.shape[0], srcs.shape[0]))
    fmm.set_source_receiver_associations(associations)

    fmm.allocate_result_arrays()  # allocate memory for Fortran arrays

    fmm.track()  # run fmst wavefront tracker code

    # collect results
    result = collect_results(options, v)

    fmm.deallocate_result_arrays()

    return result


def _calc_wavefronts_multithreading(
    v,
    recs,
    srcs,
    extent=[0.0, 1.0, 0.0, 1.0],
    options: Optional[WaveTrackerOptions] = None,
    nthreads=2,
    pool=None,
) -> WaveTrackerResult:

    # Since this function is called when there are multiple sources, we can't specify a single source for the full field calcutlation
    # Although we could create a list of source indices...
    options.ttfield_source = -1

    # Check if user provided a pool
    if pool is not None:
        created_pool = None
        executor = pool
    else:
        # Create internal ProcessPoolExecutor
        created_pool = concurrent.futures.ProcessPoolExecutor(max_workers=nthreads)
        executor = created_pool

    try:
        # Check if executor has submit method (concurrent.futures style)
        if hasattr(executor, 'submit'):
            futures = []
            for i in range(np.shape(srcs)[0]):
                futures.append(
                    executor.submit(
                        _calc_wavefronts_process, v, recs, srcs[i, :], extent, options
                    )
                )
            result_list = [f.result() for f in futures]
        else:
            # Use map for pools that don't have submit (e.g., schwimmbad pools)
            # Create a list of arguments for each source
            args_list = [(v, recs, srcs[i, :], extent, options) for i in range(np.shape(srcs)[0])]
            result_list = list(executor.map(lambda args: _calc_wavefronts_process(*args), args_list))
    finally:
        # Clean up internally created pool
        if created_pool is not None:
            created_pool.shutdown(wait=True)

    return reduce(operator.add, result_list)


def collect_results(options: WaveTrackerOptions, velocity):
    # fmst expects input spatial co-ordinates in degrees and velocities in kms/s for spherical reference frame (unless cartesian=True)
    # if cartesian is True then fmst expects input spatial co-ordinates in kms and velocities in kms/s

    ttimes = None
    raypaths = None
    frechetvals = None
    tfield = None

    if options.times:
        ttimes = fmm.get_traveltimes().copy()

    if options.paths:
        raypaths = fmm.get_raypaths().copy()

    if options.frechet:
        frechetvals = _get_frechet_derivatives(options.cartesian, options.velocityderiv, velocity)

    if options.ttfield_source >= 0:
        tfield = _get_tfield(options.cartesian, options.ttfield_source)

    return WaveTrackerResult(ttimes, raypaths, tfield, frechetvals)


def _get_frechet_derivatives(cartesian, velocityderiv, velocity):
    frechetvals = fmm.get_frechet_derivatives()

    # the frechet matrix returned in csr format and has two layers of cushion nodes surrounding the (nx,ny) grid
    F = frechetvals.toarray()  # unpack csr format
    nrays = F.shape[0]  # number of raypaths
    nx, ny = velocity.shape  # shape of non-cushion velcoity model

    # remove cushion nodes and reshape to (nx,ny)
    noncushion = _build_grid_noncushion_map(nx, ny)
    F = F[:, noncushion.flatten()].reshape((nrays, nx, ny))

    # For Spherical mode: reverse y order, for consistency (cf. ttfield array)
    if (not cartesian):
        F = F[:, :, ::-1]

    # reformat as a sparse CSR matrix
    frechetvals = csr_matrix(F.reshape((nrays, nx * ny)))

    if not velocityderiv: # change derivatives to slowness (default), otherwise stay as velocity.
        x2 = -(velocity * velocity).reshape(-1)
        frechetvals = frechetvals.multiply(x2)

    return frechetvals.copy()


def _get_tfield(cartesian, source):
    tfieldvals = fmm.get_traveltime_fields()

    tfield = tfieldvals[source].copy()

    # flip y axis of travel time field as it is provided in reverse ordered.
    if (not cartesian): 
        tfield = tfield[:, ::-1]
    return tfield


def _build_velocity_grid(v):  # maybe this could be split up a bit?
    # add cushion nodes about velocity model to be compatible with fm2dss.f90 input

    nx, ny = v.shape

    # gridc.vtx requires a single cushion layer of nodes surrounding the velocty model
    # additional boundary layer of velocities are duplicates of the nearest actual velocity value.
    vc = np.ones((nx + 2, ny + 2))
    vc[1 : nx + 1, 1 : ny + 1] = v
    vc[1 : nx + 1, 0] = v[:, 0]  # add velocities in the cushion boundary layer
    vc[1 : nx + 1, -1] = v[:, -1]  # add velocities in the cushion boundary layer
    vc[0, 1 : ny + 1] = v[0, :]  # add velocities in the cushion boundary layer
    vc[-1, 1 : ny + 1] = v[-1, :]  # add velocities in the cushion boundary layer
    vc[0, 0], vc[0, -1], vc[-1, 0], vc[-1, -1] = (
        v[0, 0],
        v[0, -1],
        v[-1, 0],
        v[-1, -1],
    )

    return vc


def _build_grid_noncushion_map(nx, ny):
    # bool array to identify cushion and non cushion nodes
    noncushion = np.zeros((nx + 2, ny + 2), dtype=bool)
    noncushion[1 : nx + 1, 1 : ny + 1] = True
    return noncushion


def _build_node_map(nx, ny):
    # mapping from cushion indices to non cushion indices
    nodemap = np.zeros((nx + 2, ny + 2), dtype=int)
    nodemap[1 : nx + 1, 1 : ny + 1] = np.array(range((nx * ny))).reshape((nx, ny))
    nodemap = nodemap[:, ::-1]
    return nodemap


def _check_sources_receivers_inside_extent(srcs, recs, extent):
    xmin = extent[0]
    xmax = extent[1]
    ymin = extent[2]
    ymax = extent[3]

    rcx = recs[:, 0]
    rcy = recs[:, 1]
    if not np.all((xmin <= rcx) & (rcx <= xmax) & (ymin <= rcy) & (rcy <= ymax)):
        raise InputError(
            msg="Input Error: One or more receiver lies outside of model extent: "
            + str(extent)
            + "\nRemedy: adjust receiver locations and run again."
        )

    srcx = srcs[:, 0]
    srcy = srcs[:, 1]
    if not np.all((xmin <= srcx) & (srcx <= xmax) & (ymin <= srcy) & (srcy <= ymax)):
        raise InputError(
            msg="Input Error: One or more source lies outside of model extent: "
            + str(extent)
            + "\nRemedy: adjust source locations and run again."
        )


def _check_requested_source_exists(tfieldsource, ns):
    if tfieldsource + 1 > ns:
        # source requested for travel time field does not exist
        print(
            f"Error: Travel time field corresponds to source: {tfieldsource}",
            "\n",
            f"      but total number of sources is {ns}.",
            "\n       No travel time field will be calculated.\n",
        )


class GridModel:  # This is for the original regular model grid (without using the basis.py package)

    def __init__(self, velocities, extent=(0, 1, 0, 1)):
        self.nx, self.ny = velocities.shape
        self.velocities = velocities
        self.xmin, self.xmax, self.ymin, self.ymax = extent
        self.xx = np.linspace(self.xmin, self.xmax, self.nx + 1)
        self.yy = np.linspace(self.ymin, self.ymax, self.ny + 1)
        self.extent = extent

    def get_velocity(self):
        return self.velocities.copy()

    def get_slowness(self):
        return 1.0 / self.velocities  # No copy needed as operation must return copy

    def set_velocity(self, v):
        assert self.velocities.shape == v.shape
        self.velocities = v.copy()

    def set_slowness(self, s):
        assert self.velocities.shape == s.shape
        self.velocities = 1.0 / s


class BasisModel:  # This is for a 2D model basis accessed through the package basis.py
    """

    A model class which is an interface to package basis.py to incorporate local or global 2D bases for tomography.

    Handles cases where model bases are local 2D pixels (default) or global 2D functions, e.g. cosine basis.
    Handles cases where input coefficients are velocities (default) or slownesses.

    Inputs:
        coeffs, ndarray(nx,ny)          : coefficients of velocity or slowness field in selected basis
        coeff_type, string              : ='velocities' then coefficients are velocities; 'slownesses' for slowness coefficients
        basis, string                   : type of model basis function
                                          `2Dpixel' for 2D regular grid of velocity/slowness values;
                                          '2Dcosine' for 2D cosine basis functions.
        ref, float                      : reference value used for perturbative representation,
                                          i.e. v[x,y] = ref + coeff[i]*basis[i,x,y]; or s[x,y] = ref + coeff[i]*basis[i,x,y], (i=1,...nx*ny)

    """

    def __init__(
        self,
        coeffs,
        extent=(0, 1, 0, 1),
        ref=0.0,
        coeff_type="velocities",
        basis="2Dpixel",
    ):
        self.nx, self.ny = coeffs.shape
        self.coeffs = coeffs
        self.xmin, self.xmax, self.ymin, self.ymax = extent
        self.extent = extent
        self.dx = (self.xmax - self.xmin) / self.nx
        self.dy = (self.ymax - self.ymin) / self.ny
        self.basis_type = basis
        self.A_calc = False
        self.vref, self.sref = 0.0, 0.0  # reference values
        if ref != 0.0:
            if coeff_type == "velocities":
                self.vref, self.sref = ref, 1.0 / ref
            else:
                self.vref, self.sref = 1.0 / ref, ref
        if self.basis_type == "2Dpixel":
            self.xx = np.linspace(self.xmin, self.xmax, self.nx + 1)
            self.yy = np.linspace(self.ymin, self.ymax, self.ny + 1)
            self.basis = base.PixelBasis2D(self.xx, self.yy)
        elif self.basis_type == "2Dcosine":
            self.basis = base.CosineBasis2D(
                self.xmin,
                self.xmax,
                self.ymin,
                self.ymax,
                self.nx,
                self.ny,
                npoints=[120, 120, 200],
            )

        self.coeff_type = coeff_type  # need to know this for non-pixel bases

    def get_velocity(self, nx=None, ny=None, returncoeff=False):
        """With no arguments this will return a velocity field.
        If bases are 2D pixels the keyword returncoeff is ignored
        If bases are not pixels and returncoeff is False, then a velocity field is evaluated and returned.
        If bases are not 2D pixels and returncoeff is True, then velocity basis coefficients are returned.
        Default values of (nx,ny) are determined by .getImage() and are the input resolution of the model.
        """
        if self.basis_type == "2Dpixel":
            if self.coeff_type == "velocities":
                if returncoeff:
                    return self.coeffs.copy()
                return self.vref + self.coeffs.copy()
            else:  # slownesses
                if returncoeff:
                    return 1.0 / self.coeffs
                return 1.0 / (self.sref + self.coeffs)
        else:  # non pixel basis
            if self.coeff_type == "velocities":
                if returncoeff:
                    return self.coeffs.copy()
                else:
                    # return a velocity field evaluated from basis summation
                    return self.vref + self.get_image(nx=nx, ny=ny)
            else:  # slownesses
                if returncoeff:
                    # coefficients are slownesses and we need to find equivalent velocity coefficients
                    return self.fit_coefficients_s2v()
                else:
                    # coefficients are slownesses and we must return a velocity field
                    return 1.0 / (self.sref + self.get_image(nx=nx, ny=ny))

    def basis_transform_matrix(self):
        if not self.A_calc:
            A = np.zeros((self.basis.nbases, self.nx * self.ny))
            for j in range(self.basis.nbases):
                A[j] = self.get_basis_image(j, nx=self.nx, ny=self.ny).flatten()
            self.A = A
            self.A_calc = True
        return self.A

    def fit_coefficientes_v2s(self, nx=None, ny=None):
        """
        calculate slowness coefficients that correspond to a given set of velocity coefficients in model basis
        """
        if nx is None:
            nx = self.nx
        if ny is None:
            ny = self.ny
        vtarget = self.get_velocity(nx=nx, ny=ny) - self.vref
        if not self.A_calc:
            A = self.basis_transform_matrix()
            self.A = A
        slowcoeff, res, rank, s = np.linalg.lstsq(
            self.A.T, 1.0 / vtarget.flatten(), rcond=None
        )  # fit slownesses coefficients to slowness field
        return slowcoeff.reshape(self.nx, self.ny)

    def fit_coefficients_s2v(self, nx=None, ny=None):
        """
        calculate velocity coefficients that correspond to a given set of slowness coefficients in model basis
        """
        if nx is None:
            nx = self.nx
        if ny is None:
            ny = self.ny
        starget = (
            self.get_slowness(nx=nx, ny=ny) - self.sref
        )  # get slowness field perturbation
        if not self.A_calc:
            A = self.basis_transform_matrix()
            self.A = A
        velcoeff, res, rank, s = np.linalg.lstsq(
            self.A.T, 1.0 / starget.flatten(), rcond=None
        )  # fit slownesses coefficients to slowness field
        return velcoeff.reshape(self.nx, self.ny)

    def convert_pixel_vel_2_basis_slow(self, v):
        """
        convert velocity model in pixel basis to equivalent model as slowness coefficients in model basis
        """
        nx, ny = v.shape
        vpert = v - self.vref
        if np.all(vpert) == 0.0:
            return np.zeros_like(v)
        # coeff = self.coeffs.copy()
        if not self.A_calc:
            A = self.basis_transform_matrix()
            self.A = A
        slowcoeff, res, rank, s = np.linalg.lstsq(
            self.A.T, 1.0 / vpert.flatten(), rcond=None
        )  # fit slownesses coefficients to slowness field
        # self.setCoeffs(coeff)
        return slowcoeff.reshape(self.nx, self.ny)

    def convert_pixel_vel_2_basis_vel(self, v):
        """
        convert velocity model in pixel basis to equivalent model as velocity coefficients in model basis
        """
        nx, ny = v.shape
        vpert = v - self.vref
        # coeff = self.coeffs.copy()
        if not self.A_calc:
            A = self.basis_transform_matrix()
            self.A = A
        velcoeff, res, rank, s = np.linalg.lstsq(
            self.A.T, vpert.flatten(), rcond=None
        )  # fit slownesses coefficients to slowness field
        # self.setCoeffs(coeff)
        return velcoeff.reshape(self.nx, self.ny)

    def get_coeffs(self):
        return self.coeffs.copy()

    def set_coeffs(self, c):
        assert self.coeffs.shape == c.shape
        self.coeffs = c

    def get_slowness(self, nx=None, ny=None, returncoeff=False):
        """
        With no arguments this will return a slowness field.
        If bases are 2D pixels the keyword returncoeff is ignored
        If bases are not pixels and returncoeff is False, then a slowness field is evaluated and returned at (nx,ny)
        If bases are not 2D pixels and returncoeff is True, then slowness basis coefficients are returned.
        Default values of (nx,ny) are determined by .getImage() and are the input resolution of the model.
        """
        if self.basis_type == "2Dpixel":
            if self.coeff_type == "velocities":
                if returncoeff:
                    return 1.0 / (self.vref + self.coeffs) - self.sref
                return 1.0 / (self.vref + self.coeffs)
            else:  # slownesses
                if returncoeff:
                    return self.coeffs.copy()
                return self.sref + self.coeffs.copy()
        else:
            if self.coeff_type == "velocities":
                if returncoeff:
                    # we need to find slowness coefficients from velocity coefficients here
                    return self.fit_coefficientes_v2s()
                else:
                    # return a slowness field after summation of velocity bases
                    return 1.0 / (self.vref + self.get_image(nx=nx, ny=ny))
            else:
                if returncoeff:
                    # coefficients are slownesses and so we return coefficients
                    return self.coeffs.copy()
                else:
                    # return a slowness field after summation of slowness bases
                    return self.sref + self.get_image(nx=nx, ny=ny)

    def set_velocity(self, v):  # set Velocity coefficients
        if self.coeff_type == "velocities":
            assert self.coeffs.shape == v.shape
            self.coeffs = v.copy()
        else:
            if self.basis_type == "2Dpixel":
                assert self.coeffs.shape == v.shape
                self.coeffs = 1.0 / (v + self.vref) - self.sref
            else:
                print(
                    " Error: can not set velocity coefficients if coeff type are slownesses and basis is not pixel"
                )
                pass  # coefficients are slownesses and we need to find and set equivalent velocity coefficients (not implemented)

    def set_slowness(self, s):  # set Slowness coefficients
        if self.coeff_type == "velocities":
            if self.basis_type == "2Dpixel":
                assert self.coeffs.shape == s.shape
                self.coeffs = 1.0 / (s + self.sref) - self.vref
            else:
                print(
                    " Error: can not set slowness coefficients if coeff type are velocities and basis is not pixel"
                )
                pass  # coefficients are velocities and we need to find and set equivalent slowness coefficients (not implemented)
        else:
            assert self.coeffs.shape == s.shape
            self.coeffs = (
                s.copy()
            )  # coefficients are slownesses and so we set new slowness coefficients

    def get_image(
        self, nx=None, ny=None
    ):  # returns 2D image of model at chosen resolution
        if nx is None:
            nx = self.nx
        if ny is None:
            ny = self.ny
        dx, dy = (self.xmax - self.xmin) / nx, (self.ymax - self.ymin) / ny
        Ym, Xm = np.meshgrid(
            np.linspace(self.ymin + dy / 2, self.ymax - dy / 2, ny),
            np.linspace(self.xmin + dx / 2.0, self.xmax - dx / 2, nx),
        )
        image = np.zeros_like(Xm)
        for j in range(
            self.basis.nbases
        ):  # sum over bases and evaluate model at each pixel in image
            image += self.coeffs.flatten()[j] * (self.basis.evaluate(j, (Xm, Ym)))
        return image

    def get_basis_image(
        self, j, nx=None, ny=None
    ):  # returns 2D image of model at chosen resolution
        if nx is None:
            nx = self.nx
        if ny is None:
            ny = self.ny
        dx, dy = (self.xmax - self.xmin) / nx, (self.ymax - self.ymin) / ny
        Ym, Xm = np.meshgrid(
            np.linspace(self.ymin + dy / 2, self.ymax - dy / 2, ny),
            np.linspace(self.xmin + dx / 2.0, self.xmax - dx / 2, nx),
        )
        return self.basis.evaluate(j, (Xm, Ym))


# --------------------------------------------------------------------------------------------
# Utility functions
# --------------------------------------------------------------------------------------------
def norm(x):
    return np.sqrt(x.dot(x))


def normalise(x):
    return x / norm(x)


def png_to_model(pngfile, nx, ny, bg=1.0, sc=1.0):
    png = Image.open(pngfile)
    png.load()
    model = sc * (
        bg
        + np.asarray(png.convert("L").resize((nx, ny)).transpose(Image.ROTATE_270))
        / 255.0
    )
    return model


def generate_surface_points(
    nPerSide,
    extent=(0, 1, 0, 1),
    surface=[True, True, True, True],
    addCorners=True,
):
    out = []
    x = np.linspace(extent[0], extent[1], nPerSide + 2)[1 : nPerSide + 1]
    y = np.linspace(extent[2], extent[3], nPerSide + 2)[1 : nPerSide + 1]
    if surface[0]:
        out += [[extent[0], _y] for _y in y]
    if surface[1]:
        out += [[extent[1], _y] for _y in y]
    if surface[2]:
        out += [[_x, extent[2]] for _x in x]
    if surface[3]:
        out += [[_x, extent[3]] for _x in x]
    if addCorners:
        if surface[0] or surface[2]:
            out += [[extent[0], extent[2]]]
        if surface[0] or surface[3]:
            out += [[extent[0], extent[3]]]
        if surface[1] or surface[2]:
            out += [[extent[1], extent[2]]]
        if surface[1] or surface[3]:
            out += [[extent[1], extent[3]]]
    return np.array(out)


# --------------------------------------------------------------------------------------------
# Plotting routines
# --------------------------------------------------------------------------------------------

def display_model_orig(
    model,
    paths=None,
    extent=(0, 1, 0, 1),
    clim=None,
    cmap=None,
    figsize=(6, 6),
    title=None,
    line=1.0,
    cline="k",
    alpha=1.0,
    points=None,
    wfront=None,
    cwfront="k",
    diced=True,
    dicex=8,
    dicey=8,
    cbarshrink=0.6,
    cbar=True,
    filename=None,
    reversedepth=False,
    points_size = 1.0,
    aspect = None,
    **wkwargs,
):
    """

    Function to plot 2D velocity or slowness field

    Inputs:
        model, ndarray(nx,ny)           : 2D velocity or slowness field on rectangular grid
        paths, string                   :

    """

    fig = plt.figure(figsize=figsize)
    
    if cmap is None:
        cmap = plt.cm.RdBu

    # if diced option plot the actual B-spline interpolated velocity used by fmst program

    plotmodel = model
    if diced:
        plotmodel = create_diced_grid(model, extent=extent, dicex=dicex, dicey=dicey)
    
    if(reversedepth):
        extentr = [extent[0],extent[1],extent[3],extent[2]]
        plt.imshow(plotmodel.T, origin="upper", extent=extentr, aspect=aspect, cmap=cmap)
    else:
        plt.imshow(plotmodel.T, origin="lower", extent=extent, aspect=aspect, cmap=cmap)
        
    if paths is not None:
        if isinstance(paths, np.ndarray) and paths.ndim == 2:
            if paths.shape[1] == 4:  # we have paths from xrt.tracer so adjust
                paths = change_paths_format(paths)

        for i in range(len(paths)):
            p = paths[i]
            if(type(cline) is list):
                cl = cline[i]
            else:
                cl = cline
            if(type(line) is list):
                lw = line[i]
            else:
                lw = line
            plt.plot(p[:, 0], p[:, 1], cl, lw=lw, alpha=alpha)

    if clim is not None:
        plt.clim(clim)

    if title is not None:
        plt.title(title)

    if wfront is not None:
        nx, ny = wfront.shape
        X, Y = np.meshgrid(
            np.linspace(extent[0], extent[1], nx),
            np.linspace(extent[2], extent[3], ny),
        )
        if(False):
            plt.contour(X, Y, wfront.T[::-1], **wkwargs)  # Negative contours default to dashed.
        else:
            plt.contour(X, Y, wfront.T, **wkwargs)  # Negative contours default to dashed.

    if wfront is None and cbar:
        plt.colorbar(shrink=cbarshrink)

    if points is not None:
        plt.plot(points[:, 0], points[:, 1], 'bo',markersize=points_size)

    if(reversedepth):
        plt.xlim(extent[0],extent[1])
        plt.ylim(extent[3],extent[2])
    else:
        plt.xlim(extent[0],extent[1])
        plt.ylim(extent[2],extent[3])
    
    if filename is not None:
        plt.savefig(filename)

    plt.show()

def display_model(
    model,
    paths=None,
    extent=(0, 1, 0, 1),
    clim=None,
    cmap=None,
    figsize=(6, 6),
    title=None,
    line=1.0,
    cline="k",
    alpha=1.0,
    points=None,
    wfront=None,
    cwfront="k",
    diced=True,
    dicex=8,
    dicey=8,
    cbarshrink=0.6,
    cbar=True,
    filename=None,
    reversedepth=False,
    points_size = 1.0,
    aspect = None,
    ax=None,  # <-- NEW: Optional Axes object
    **wkwargs,
):
    """
    Function to plot 2D velocity or slowness field.
    Can plot onto an existing axis if 'ax' is provided.

    Inputs:
        model, ndarray(nx,ny)          : 2D velocity or slowness field on rectangular grid
        paths, string                  : ...
        ax, matplotlib.axes.Axes       : Optional axis object to plot onto.
    """

    # --- Setup Figure and Axes ---
    if ax is None:
        # If no axis is provided, create a new figure and axis
        fig, ax = plt.subplots(figsize=figsize)
    else:
        # If an axis is provided, make sure we don't call plt.figure() or plt.show()
        fig = ax.figure # Retain a reference to the figure if needed later

    if cmap is None:
        cmap = plt.cm.RdBu

    # --- Prepare Data ---
    plotmodel = model
    if diced:
        plotmodel = create_diced_grid(model, extent=extent, dicex=dicex, dicey=dicey)

    # --- Plot Image (imshow) ---
    if reversedepth:
        extentr = [extent[0], extent[1], extent[3], extent[2]]
        # Use ax.imshow() instead of plt.imshow()
        im = ax.imshow(
            plotmodel.T, origin="upper", extent=extentr, aspect=aspect, cmap=cmap
        )
    else:
        # Use ax.imshow() instead of plt.imshow()
        im = ax.imshow(
            plotmodel.T, origin="lower", extent=extent, aspect=aspect, cmap=cmap
        )

    # --- Plot Paths (plot) ---
    if paths is not None:
        if isinstance(paths, np.ndarray) and paths.ndim == 2:
            if paths.shape[1] == 4:  # we have paths from xrt.tracer so adjust
                paths = change_paths_format(paths)

        for i in range(len(paths)):
            p = paths[i]
            if type(cline) is list:
                cl = cline[i]
            else:
                cl = cline
            if type(line) is list:
                lw = line[i]
            else:
                lw = line
            # Use ax.plot() instead of plt.plot()
            ax.plot(p[:, 0], p[:, 1], cl, lw=lw, alpha=alpha)

    # --- Set Color Limits (clim) ---
    if clim is not None:
        # We can set the clim directly on the image object (im)
        im.set_clim(clim)

    # --- Set Title (title) ---
    if title is not None:
        # Use ax.set_title() instead of plt.title()
        ax.set_title(title)

    # --- Plot Wavefronts (contour) ---
    if wfront is not None:
        nx, ny = wfront.shape
        X, Y = np.meshgrid(
            np.linspace(extent[0], extent[1], nx),
            np.linspace(extent[2], extent[3], ny),
        )
        # Use ax.contour() instead of plt.contour()
        if False: # Retaining original conditional for contour
             ax.contour(X, Y, wfront.T[::-1], **wkwargs)
        else:
             ax.contour(X, Y, wfront.T, **wkwargs)

    # --- Add Colorbar (colorbar) ---
    if wfront is None and cbar:
        # Use fig.colorbar() and pass the image (im) and axis (ax)
        fig.colorbar(im, ax=ax, shrink=cbarshrink)

    # --- Plot Points (plot) ---
    if points is not None:
        # Use ax.plot() instead of plt.plot()
        ax.plot(points[:, 0], points[:, 1], 'bo', markersize=points_size)

    # --- Set Axis Limits (xlim/ylim) ---
    if reversedepth:
        # Use ax.set_xlim() and ax.set_ylim()
        ax.set_xlim(extent[0], extent[1])
        ax.set_ylim(extent[3], extent[2])
    else:
        # Use ax.set_xlim() and ax.set_ylim()
        ax.set_xlim(extent[0], extent[1])
        ax.set_ylim(extent[2], extent[3])
    
    # --- Save and Show (Only if we created the figure) ---
    # These should only be called if we created the figure (ax is None initially)
    if ax is fig.axes[0]: # Simple check if this is the only (first) axis on the fig
        if filename is not None:
            fig.savefig(filename)
            
    if(ax is None): plt.show() # Only call show if we created the figure

    return

def create_diced_grid(v, extent=[0.0, 1.0, 0.0, 1.0], dicex=8, dicey=8):
    nx, ny = v.shape
    x = np.linspace(extent[0], extent[1], nx)
    y = np.linspace(extent[2], extent[3], ny)
    kx, ky = 3, 3
    if nx <= 3:
        kx = nx - 1  # reduce order of B-spline if we have too few velocity nodes
    if ny <= 3:
        ky = ny - 1
    rect = RectBivariateSpline(x, y, v, kx=kx, ky=ky)
    xx = np.linspace(extent[0], extent[1], dicex * nx)
    yy = np.linspace(extent[2], extent[3], dicey * ny)
    X, Y = np.meshgrid(xx, yy, indexing="ij")
    vinterp = rect.ev(X, Y)
    return vinterp


def change_paths_format(paths):
    p = np.zeros((len(paths), 2, 2))
    for i in range(len(paths)):
        p[i, 0, 0] = paths[i, 0]
        p[i, 0, 1] = paths[i, 1]
        p[i, 1, 0] = paths[i, 2]
        p[i, 1, 1] = paths[i, 3]
    return p



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 13 07:31:45 2021

@author: Malcolm Sambridge
"""
import numpy as np
import pickle

# Define available basis classes


class PixelBasis2D:  # Model basis class
    """
    A 2D Model pixel basis object.
    A class that is used by various routines to calculate Jacobians through numerical integration.

    """

    def __init__(
        self, xg, zg, npoints=[50, 50, 1000], outer="x"
    ):  # receives linear grid dimension
        self.type = "2Dpixel"
        self.name = "2Dpixel"
        self.xg = xg  # X-grid for voxel model
        self.zg = zg  # Y-grid for voxel model
        self.nx = len(xg) - 1  # Number of cells in X-direction
        self.nz = len(zg) - 1  # Number of cells in Y-direction
        self.xmin = xg[0]  # X-min of model
        self.xmax = xg[-1]  # X-max of model
        self.zmin = zg[0]  # Y-min of model
        self.zmax = zg[-1]  # Y-max of model
        self.outer = outer  # axis for outer loop of basis grid (for indexing)
        self.nbases = self.nx * self.nz  # Number of basis functions in voxel model

        # X-integration grid used in each cell of this basis
        # (only used if kernel is 3D and integration mthd is 'simpson')
        self.nxi = npoints[0]
        # Y-integration grid used in each cell of this basis
        # (only used if kernel is 3D and integration mthd is 'simpson')
        self.nzi = npoints[1]
        # Number of integration points along 1D ray within cell
        # (only used if kernel is a 1D and integration method `simpson')
        self.nline = npoints[2]

    def evaluate(self, j, pos):
        """evaluate the ith model basis function at location (x,y)"""
        if j < 0 or j >= self.nbases:
            raise BasisError(j)
        x, z = pos
        ix, iz = self.convert_index(j)  # convert j basis to 2D model basis index
        if np.isscalar(x):
            b = 0.0  # pixel basis value
            if (
                (x >= self.xg[ix])
                & (x < self.xg[ix + 1])
                & (z >= self.zg[iz])
                & (z < self.zg[iz + 1])
            ):
                b = 1.0
            if (ix == self.nx - 1 and x == self.xg[-1]) or (
                iz == self.nz - 1 and z == self.zg[-1]
            ):
                b = 1.0  # Correct upper boundary
        else:
            b = np.zeros_like(x)
            b[
                (x >= self.xg[ix])
                & (x < self.xg[ix + 1])
                & (z >= self.zg[iz])
                & (z < self.zg[iz + 1])
            ] = 1.0
            b[(ix == self.nx - 1 and x == self.xg[-1])] = 1.0
            b[(iz == self.nz - 1 and z == self.zg[-1])] = 1.0
        return b

    def convert_index(self, i):
        """
        Convert single 1D array index to 2D array index using inner loop over X and outer loop over Y
        """
        if i < 0 or i >= self.nbases:
            raise BasisError(i)
        if self.outer == "x":
            zindex = int(i % self.nz)
            xindex = int((i / self.nz) % self.nx)
        else:
            xindex = int(i % self.nx)
            zindex = int((i / self.nx) % self.nz)
        return xindex, zindex

    def iconvert_index(self, ix, iz):
        """
        Convert single 2D array index to 1D array index inner loop over X and outer loop over Y
        """
        if ix < 0 or ix >= self.nx:
            raise BasisError(ix)
        if iz < 0 or iz >= self.nz:
            raise BasisError(iz)
        if self.outer == "x":
            return ix * self.nz + iz  # iz is the inner loop and ix the outer loop
        else:
            return iz * self.nx + ix  # ix is the inner loop and iz the outer loop

    def int_limits(self, j):
        """calculates the limits of integration for given basis function"""
        ix, iz = self.convert_index(j)  # convert j basis to 2D model basis index
        # For voxel basis we return the limits of the jth cell (because jth basis is zero elsewhere)
        return [self.xg[ix], self.xg[ix + 1]], [self.zg[iz], self.zg[iz + 1]]


class BasisError(Exception):
    """Raised when input model basis id does not exist in kernel definition"""

    def __init__(self, cset=[]):
        super().__init__("\n Basis id " + str(cset) + " not recognized. \n")


class CosineBasis2D:  # Model basis class
    """
    A 2D Model cosine basis object.
    A class that is used by various routines to calculate Jacobians through numerical integration.

    """

    def __init__(
        self, x0, x1, z0, z1, nx, ny, npoints=[120, 120, 200]
    ):  # receives linear grid dimension
        self.type = "2D"
        self.name = "2Dcosine"
        self.xmin = x0  # X-min of model
        self.xmax = x1  # X-max of model
        self.zmin = z0  # Z-min of model
        self.zmax = z1  # Z-max of model
        self.nx = nx  # set number of bases in X-direction
        self.nz = ny  # set number of bases in Z-direction
        self.nbases = self.nx * self.nz  # Total number of basis functions in 2D model

        self.nxi = npoints[0]
        # X-integration grid used in each cell of this basis
        # (only used if kernel is 3D and integration mthd is 'simpson')
        self.nzi = npoints[1]
        # Y-integration grid used in each cell of this basis
        # (only used if kernel is 3D and integration mthd is 'simpson')
        self.nline = npoints[2]
        # Number of integration points along 1D ray within cell
        # (only used if kernel is a 1D and integration method `simpson')
        self.Lx = self.xmax - self.xmin  # Set maximum X wavelength in model
        self.Lz = self.zmax - self.zmin  # Set maximum Z wavelength in model
        self.norm = np.sqrt(self.Lx * self.Lz)
        self.area = self.Lx * self.Lz  # set area of domain

    def evaluate(self, j, pos):  # evaluate the ith data kernel at location (x,z)
        x, z = pos
        ix, iz = self.convert_index(
            j
        )  # convert from single index to pair of ix,iz indices
        b = np.cos(np.pi * ix * (x - self.xmin) / (self.Lx)) * np.cos(
            np.pi * iz * (z - self.zmin) / (self.Lz)
        )  # evaluate jth basis function at input positions
        fx, fz = np.sqrt(2.0), np.sqrt(2.0)
        if ix == 0:
            fx = 1.0
        if iz == 0:
            fz = 1.0
        return b * fx * fz / self.norm

    def convert_index(self, i):
        """
        Convert single 1D array index to 2D array index using inner loop over X and outer loop over Y
        """
        if i < 0 or i >= self.nbases:
            raise BasisError(i)

        xindex = int(i % self.nx)
        zindex = int((i / self.nx) % self.nz)
        return xindex, zindex

    def iconvert_index(self, ix, iz):
        """
        Convert single 2D array index to 1D array index inner loop over X and outer loop over Y
        """
        if ix < 0 or ix >= self.nx:
            raise BasisError(ix)
        if iz < 0 or iz >= self.nz:
            raise BasisError(iz)
        return iz * self.nx + ix  # ix is the inner loop and iz the outer loop

    def int_limits(self, j):
        """calculates the limits of integration for given basis function"""
        return [self.xmin, self.xmax], [
            self.zmin,
            self.zmax,
        ]  # Here we choose whole model

    def fit_coefficients(self, target, A=None, returnmatrix=False):
        """
        fitcoefficients - fits basis coefficients to given field of function values at arbitrary resolution.

        Inputs:
             target, ndarray(nx,nz) -               : input function field to fit model basis to
             A, ndarray(nx*nz,self.nx*self.nz) -    : transformation matrix of basis coefficients at each point in inpit field.
                                                      A[i,j] is the j basis function evaluated at ith point in the function field,
                                                      where both indices correspond to flattened arrays.
                                                      A is calculated if A=None, and returned if returnmaxtrix = True.
        Returns:
            coeff, ndarray(self.nx,self.ny)         : best fit basis coefficients
        """
        nx, nz = target.shape
        if type(A) is not np.ndarray:
            A = np.zeros((self.nbases, nx * nz))
            for j in range(self.nbases):
                A[j] = self.get_basis_image(j, nx=nx, nz=nz).flatten()
        coeff, res, rank, s = np.linalg.lstsq(
            A.T, target.flatten(), rcond=None
        )  # fit basis coefficients to input field
        if returnmatrix:
            return coeff.reshape(self.nx, self.nz), A
        return coeff.reshape(self.nx, self.nz)

    def get_basis_image(self, j, nx=None, nz=None):
        """returns 2D image of model at chosen resolution"""
        if nx is None:
            nx = self.nx
        if nz is None:
            nz = self.nz
        dx, dz = (self.xmax - self.xmin) / nx, (self.zmax - self.zmin) / nz
        Zm, Xm = np.meshgrid(
            np.linspace(self.zmin + dz / 2, self.zmax - dz / 2, nz),
            np.linspace(self.xmin + dx / 2.0, self.xmax - dx / 2, nx),
        )
        return self.evaluate(j, (Xm, Zm))

    # Define available data kernel classes


class RayKernel1D:  # A linear data kernel class
    """
    Data kernel object.

    A class that is used by various routines to calculate Jacobians through numerical integration.

    """

    def __init__(self, paths):  # receives the end points of rays.
        self.nkernels = len(paths)
        self.paths = paths
        self.constant = 1.0  # kernel is a constant along the ray (typically 1.0)
        d = np.zeros(self.nkernels)
        if isinstance(paths, list):
            self.type = "1Dcurveray"
            dcumsum = []
            for k, p in enumerate(paths):
                dl = p[1:] - p[:-1]  # co-ordinates of ray segment end points
                dls = np.linalg.norm(dl, axis=1)  # lengths of each ray segment
                dcumsum.append(
                    np.array([0.0, *np.cumsum(dls)])
                )  # cummulative lengths of each ray segment
                d[k] = np.sum(dls)  # lengths of each ray
            self.cumlengths = dcumsum
        else:
            self.type = "1Dstraightray"  # This id indicates a straight ray (use `1D' if ray is not straight line)
            for k, p in enumerate(paths):
                dl = p[1:] - p[:-1]  # co-ordinates of ray segment end points
                dls = np.linalg.norm(dl, axis=1)  # lengths of each ray segment
                d[k] = np.sum(dls)  # lengths of each ray
        self.lengths = d

    def evaluate(self, i):  # evaluate the ith data kernel at location (x,y,z)
        if i < 0 or i >= self.nkernels:
            raise KernelOBSError(i)

        return self.constant  # Return constant along ray for integration

    def position(self, i, l):
        """
        find (x,y,z) position of ith data kernel at length l along ray.

        This example is for a straight seismic ray (self.type = '1Dstraightray').
        For a curved ray (self.type = '1D') this is where one would specify (x,y,z) as a function of length l.
        """
        if i < 0 or i >= self.nkernels:
            raise KernelOBSError(i)
        # if(l <= 0): return self.paths[i][0]
        # if(l >= self.lengths[i]): return self.paths[i][-1]
        if self.type == "1Dstraightray":
            alpha = l / self.lengths[i]
            a0 = self.paths[i][0]
            a1 = self.paths[i][1]
            return (1.0 - alpha) * a0[:, np.newaxis] + alpha * a1[
                :, np.newaxis
            ]  # here we use a straight line between endpoints of locations
        else:
            j = np.clip(
                np.searchsorted(self.cumlengths[i], l), 1, len(self.cumlengths[i]) - 1
            )
            # print(i,self.lengths[i],l,j,self.cumlengths[i])
            d0 = self.cumlengths[i][j - 1]
            d1 = self.cumlengths[i][j]
            dl = d1 - d0
            a0 = self.paths[i][j - 1]
            a1 = self.paths[i][j]
            alpha = (l - d0) / dl
            out = (1.0 - alpha[:, np.newaxis]) * a0 + alpha[
                :, np.newaxis
            ] * a1  # interpolate along segment for location
            return out.T


class Error(Exception):
    """Base class for other exceptions"""

    pass


class KernelOBSError(Exception):
    """Raised when input kernel id does not exist in kernel definition"""

    def __init__(self, cset=[]):
        super().__init__("\n Kernel id " + str(cset) + " not recognized. \n")


# Define utility routines for 2D tomography


def randomrays2Dbox(Nrays, x, y, seed=61254557, filename=""):
    """Generate random rays across 2D model"""
    if filename == "":  # generate ray endpoints randomly on side of given box
        np.random.seed(seed)
        sides = np.zeros((Nrays, 2), dtype=int)
        for i in range(Nrays):
            sides[i] = np.random.choice(4, replace=False, size=2)
        paths2D = np.zeros((Nrays, 2, 2))
        for i in range(Nrays):
            paths2D[i, 0] = _choose_endpoint(sides[i, 0], x, y)
            paths2D[i, 1] = _choose_endpoint(sides[i, 1], x, y)
    else:  # read ray end points from pickle file
        with open(filename, "rb") as f:
            paths2D = pickle.load(f)
    return paths2D


def _choose_endpoint(side, x, y):  # choose random point on boundary of 2D box
    if side == 0:
        # choose endpoint randomly along xmin side
        return [
            x[0],
            y[0] + (y[-1] - y[0]) * np.random.random(),
        ]
    elif side == 1:
        # choose endpoint randomly along xmax side
        return [
            x[-1],
            y[0] + (y[-1] - y[0]) * np.random.random(),
        ]
    elif side == 2:
        # choose endpoint randomly along ymin side
        return [
            x[0] + (x[-1] - x[0]) * np.random.random(),
            y[0],
        ]
    else:
        # choose endpoint randomly along ymax side
        return [
            x[0] + (x[-1] - x[0]) * np.random.random(),
            y[-1],
        ]


def _generate_center_coordinates(l_x):
    l_x = float(l_x)
    X, Y = np.mgrid[:l_x, :l_x]
    center = l_x / 2.0
    X += 0.5 - center
    Y += 0.5 - center
    return X, Y



import numpy as np
import scipy
import ctypes
from pathlib import Path
from site import getsitepackages


package_dir = Path(getsitepackages()[0]) / "pyfm2d"
lib_path = next(package_dir.glob("libfm2d*.so"))
libfm2d = ctypes.cdll.LoadLibrary(str(lib_path))


def fmmin2d():
    libfm2d.fmmin2d()


def track():
    libfm2d.track()


def read_solver_options(fn_):
    fn = ctypes.c_char_p(fn_.encode("UTF-8"))
    libfm2d.read_configuration(fn, ctypes.c_int(len(fn.value)))


def set_solver_options(
    gdx, gdz, asgr, sgdl, sgs, earth, fom, snb, fsrt, cfd, wttf, wrgf, cart, quiet,
):
    # Not sure if the np.int32 and np.float32 are necessary
    # e.g. ctypes.c_int is an alias for ctypes.c_long anyway on 64-bit systems
    # so casting to 32-bit first seems unnecessary
    #
    # unit test passed on my system without the casting
    gdx_ = ctypes.c_int(np.int32(gdx))
    gdz_ = ctypes.c_int(np.int32(gdz))
    asgr_ = ctypes.c_int(np.int32(asgr))
    sgdl_ = ctypes.c_int(np.int32(sgdl))
    sgs_ = ctypes.c_int(np.int32(sgs))
    earth_ = ctypes.c_float(np.float32(earth))
    fom_ = ctypes.c_int(np.int32(fom))
    snb_ = ctypes.c_float(np.float32(snb))

    fsrt_ = ctypes.c_int(np.int32(fsrt))
    cfd_ = ctypes.c_int(np.int32(cfd))
    wttf_ = ctypes.c_int(np.int32(wttf))
    wrgf_ = ctypes.c_int(np.int32(wrgf))
    cart_ = ctypes.c_int(np.int32(cart))
    quiet_ = ctypes.c_int(np.int32(quiet))

    libfm2d.set_solver_options(
        ctypes.byref(gdx_),
        ctypes.byref(gdz_),
        ctypes.byref(asgr_),
        ctypes.byref(sgdl_),
        ctypes.byref(sgs_),
        ctypes.byref(earth_),
        ctypes.byref(fom_),
        ctypes.byref(snb_),
        ctypes.byref(fsrt_),
        ctypes.byref(cfd_),
        ctypes.byref(wttf_),
        ctypes.byref(wrgf_),
        ctypes.byref(cart_),
        ctypes.byref(quiet_),
    )


def get_solver_options():
    gdx_ = ctypes.c_int(-99)
    gdz_ = ctypes.c_int(-99)
    asgr_ = ctypes.c_int(-99)
    sgdl_ = ctypes.c_int(-99)
    sgs_ = ctypes.c_int(-99)
    earth_ = ctypes.c_float(-99.9)
    fom_ = ctypes.c_int(-99)
    snb_ = ctypes.c_float(-99)

    fsrt_ = ctypes.c_int(-99)
    cfd_ = ctypes.c_int(-99)
    wttf_ = ctypes.c_int(-99)
    wrgf_ = ctypes.c_int(-99)
    cart_ = ctypes.c_int(-99)
    quiet_ = ctypes.c_int(-99)

    libfm2d.get_solver_options(
        ctypes.byref(gdx_),
        ctypes.byref(gdz_),
        ctypes.byref(asgr_),
        ctypes.byref(sgdl_),
        ctypes.byref(sgs_),
        ctypes.byref(earth_),
        ctypes.byref(fom_),
        ctypes.byref(snb_),
        ctypes.byref(fsrt_),
        ctypes.byref(cfd_),
        ctypes.byref(wttf_),
        ctypes.byref(wrgf_),
        ctypes.byref(cart_),
        ctypes.byref(quiet_),
    )

    gdx = gdx_.value
    gdz = gdz_.value
    asgr = asgr_.value
    sgdl = sgdl_.value
    sgs = sgs_.value
    earth = earth_.value
    fom = fom_.value
    snb = snb_.value

    fsrt = fsrt_.value
    cfd = cfd_.value
    wttf = wttf_.value
    wrgf = wrgf_.value
    cart = cart_.value
    quiet = quiet_.value

    return gdx, gdz, asgr, sgdl, sgs, earth, fom, snb, fsrt, cfd, wttf, wrgf, cart, quiet


def read_sources(fn_):
    fn = ctypes.c_char_p(fn_.encode("UTF-8"))
    libfm2d.read_sources(fn, ctypes.c_int(len(fn.value)))


def set_sources(scx_, scz_):
    # testing showed that dtype must be np.float32
    # not sure why that's the case here but not for the options
    # must be something to do with the memory allocation in the
    # Fortran code
    scx_ = np.asfortranarray(scx_, dtype=np.float32)
    scz_ = np.asfortranarray(scz_, dtype=np.float32)
    nsrc_ = ctypes.c_int(scx_.shape[0])

    libfm2d.set_sources(
        scx_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        scz_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        ctypes.byref(nsrc_),
    )


def get_sources():
    nsrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_sources(ctypes.byref(nsrc_))
    nsrc = nsrc_.value
    scx_ = np.empty((nsrc), dtype=ctypes.c_float)
    scz_ = np.empty((nsrc), dtype=ctypes.c_float)
    libfm2d.get_sources(
        scx_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        scz_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        ctypes.byref(nsrc_),
    )
    scx = np.array(scx_)
    scz = np.array(scz_)
    return scx, scz


def read_receivers(fn_):
    fn = ctypes.c_char_p(fn_.encode("UTF-8"))
    libfm2d.read_receivers(fn, ctypes.c_int(len(fn.value)))


def set_receivers(rcx_, rcz_):
    rcx_ = np.asfortranarray(rcx_, dtype=np.float32)
    rcz_ = np.asfortranarray(rcz_, dtype=np.float32)
    nrc_ = ctypes.c_int(rcx_.shape[0])

    libfm2d.set_receivers(
        rcx_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        rcz_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        ctypes.byref(nrc_),
    )


def get_receivers():
    nrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_receivers(ctypes.byref(nrc_))
    nrc = nrc_.value
    rcx_ = np.empty((nrc), dtype=ctypes.c_float)
    rcz_ = np.empty((nrc), dtype=ctypes.c_float)
    libfm2d.get_receivers(
        rcx_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        rcz_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        ctypes.byref(nrc_),
    )
    rcx = np.array(rcx_)
    rcz = np.array(rcz_)
    return rcx, rcz


def read_source_receiver_associations(fn_):
    fn = ctypes.c_char_p(fn_.encode("UTF-8"))
    libfm2d.read_source_receiver_associations(fn, ctypes.c_int(len(fn.value)))


def set_source_receiver_associations(srs):
    nsrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_sources(ctypes.byref(nsrc_))
    nrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_receivers(ctypes.byref(nrc_))

    srs_ = np.asfortranarray(srs, dtype=np.int32)
    srs_ = libfm2d.set_source_receiver_associations(
        srs_.ctypes.data_as(ctypes.POINTER(ctypes.c_int))
    )

    return


def get_source_receiver_associations():

    nsrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_sources(ctypes.byref(nsrc_))
    nsrc = nsrc_.value

    nrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_receivers(ctypes.byref(nrc_))
    nrc = nrc_.value

    srs_ = np.asfortranarray(np.zeros([nrc, nsrc]), dtype=np.int32)
    libfm2d.get_source_receiver_associations(
        srs_.ctypes.data_as(ctypes.POINTER(ctypes.c_int))
    )

    srs = np.array(srs_)
    return srs


def read_velocity_model(fn_):
    fn = ctypes.c_char_p(fn_.encode("UTF-8"))
    libfm2d.read_velocity_model(fn, ctypes.c_int(len(fn.value)))


def set_velocity_model(nvx, nvz, goxd, gozd, dvxd, dvzd, velv):
    nvx_ = ctypes.c_int(np.int32(nvx))
    nvz_ = ctypes.c_int(np.int32(nvz))
    goxd_ = ctypes.c_float(np.float32(goxd))
    gozd_ = ctypes.c_float(np.float32(gozd))
    dvxd_ = ctypes.c_float(np.float32(dvxd))
    dvzd_ = ctypes.c_float(np.float32(dvzd))
    velv_ = np.asfortranarray(velv, dtype=np.float32)

    libfm2d.set_velocity_model(
        ctypes.byref(nvx_),
        ctypes.byref(nvz_),
        ctypes.byref(goxd_),
        ctypes.byref(gozd_),
        ctypes.byref(dvxd_),
        ctypes.byref(dvzd_),
        velv_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
    )


def get_velocity_model():
    nvx_ = ctypes.c_int(-99)
    nvz_ = ctypes.c_int(-99)

    libfm2d.get_number_of_velocity_model_vertices(
        ctypes.byref(nvx_), ctypes.byref(nvz_)
    )

    goxd_ = ctypes.c_float(-99.9)
    gozd_ = ctypes.c_float(-99.9)

    dvxd_ = ctypes.c_float(-99.9)
    dvzd_ = ctypes.c_float(-99.9)

    velv_ = np.asfortranarray(
        np.zeros([nvz_.value + 2, nvx_.value + 2]), dtype=np.float32
    )

    libfm2d.get_velocity_model(
        ctypes.byref(nvx_),
        ctypes.byref(nvz_),
        ctypes.byref(goxd_),
        ctypes.byref(gozd_),
        ctypes.byref(dvxd_),
        ctypes.byref(dvzd_),
        velv_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
    )

    nvx = nvx_.value
    nvz = nvz_.value

    goxd = goxd_.value
    gozd = gozd_.value

    dvxd = dvxd_.value
    dvzd = dvzd_.value

    return nvx, nvz, goxd, gozd, dvxd, dvzd, velv_


def allocate_result_arrays():
    libfm2d.allocate_result_arrays()


def deallocate_result_arrays():
    libfm2d.deallocate_result_arrays()


def get_traveltimes():
    nttimes_ = ctypes.c_int(-99)
    libfm2d.get_number_of_traveltimes(ctypes.byref(nttimes_))
    nttimes = nttimes_.value

    ttimes_ = np.asfortranarray(np.zeros(nttimes), dtype=np.float32)

    tids_ = np.asfortranarray(np.zeros(nttimes), dtype=np.int32)

    libfm2d.get_traveltimes(
        ttimes_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        tids_.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
    )

    ttimes = np.array(ttimes_)

    return ttimes


def get_frechet_derivatives():

    frechet_nnz_ = ctypes.c_int(-99)
    libfm2d.get_number_of_frechet_derivatives(ctypes.byref(frechet_nnz_))
    frechet_nnz = frechet_nnz_.value

    frechet_irow_ = np.asfortranarray(np.zeros(frechet_nnz), dtype=np.int32)
    frechet_icol_ = np.asfortranarray(np.zeros(frechet_nnz), dtype=np.int32)
    frechet_val_ = np.asfortranarray(np.zeros(frechet_nnz), dtype=np.float32)

    libfm2d.get_frechet_derivatives(
        frechet_irow_.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        frechet_icol_.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        frechet_val_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
    )

    jrow = np.array(frechet_irow_) - 1
    jcol = np.array(frechet_icol_) - 1
    jval = np.array(frechet_val_)

    nvx_ = ctypes.c_int(-99)
    nvz_ = ctypes.c_int(-99)

    libfm2d.get_number_of_velocity_model_vertices(
        ctypes.byref(nvx_), ctypes.byref(nvz_)
    )
    nvx = nvx_.value
    nvz = nvz_.value

    nsrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_sources(ctypes.byref(nsrc_))
    nsrc = nsrc_.value
    nrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_receivers(ctypes.byref(nrc_))
    nrc = nrc_.value

    return scipy.sparse.csr_array(
        (jval, (jrow, jcol)), shape=(nsrc * nrc, (nvx + 2) * (nvz + 2))
    )


def get_raypaths():

    npaths_ = ctypes.c_int(-99)
    libfm2d.get_number_of_raypaths(ctypes.byref(npaths_))
    npaths = int(npaths_.value)

    max_nppts_ = ctypes.c_int(-99)
    libfm2d.get_maximum_number_of_points_per_raypath(ctypes.byref(max_nppts_))
    max_nppts = int(max_nppts_.value)

    paths_ = np.asfortranarray(np.zeros([npaths, max_nppts, 2]), dtype=np.float32)

    nppts_ = np.asfortranarray(np.zeros([npaths]), dtype=np.int32)

    libfm2d.get_raypaths(
        paths_.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        nppts_.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
    )

    paths = []

    for i in range(npaths):
        lat = paths_[i, 0 : nppts_[i], 0]
        long = paths_[i, 0 : nppts_[i], 1]
        path = np.array([long, lat]).T
        paths.append(path)

    return paths


def get_traveltime_fields():
    nsrc_ = ctypes.c_int(-99)
    libfm2d.get_number_of_sources(ctypes.byref(nsrc_))
    nsrc = nsrc_.value

    nnx_ = ctypes.c_int(-99)
    nnz_ = ctypes.c_int(-99)

    libfm2d.get_number_of_grid_nodes(ctypes.byref(nnx_), ctypes.byref(nnz_))
    nnx = nnx_.value
    nnz = nnz_.value

    tfields_ = np.asfortranarray(np.zeros([nsrc, nnz, nnx]), dtype=np.float32)

    libfm2d.get_traveltime_fields(
        tfields_.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    )

    tfields = np.array(tfields_)
    return tfields



import os
import numpy as np
import pytest
from dataclasses import dataclass

from pathlib import Path

from pyfm2d import fastmarching as fmm

DATADIR = Path(__file__).parent
FMINFILE = str(DATADIR / "fm2dss.in")
VELMODELFILE = str(DATADIR / "gridc.vtx")
SOURCESFILE = str(DATADIR / "sources.dat")
RECEIVERSFILE = str(DATADIR / "receivers.dat")
ASSOCIATIONSFILE = str(DATADIR / "otimes.dat")


@dataclass
class FMMOptions:
    dicex: int = 8
    dicey: int = 8
    sourcegridrefine: int = 1
    sourcedicelevel: int = 5
    sourcegridsize: int = 10
    earthradius: float = 6371.0
    schemeorder: int = 1
    nbsize: float = 0.5
    lttimes: int = 1
    lfrechet: int = 1
    tfieldsource: int = 0
    lpaths: int = -1
    lcartesian: int = 0


def test_set_solver_options():
    options = FMMOptions()

    fmm.set_solver_options(
        options.dicex,
        options.dicey,
        options.sourcegridrefine,
        options.sourcedicelevel,
        options.sourcegridsize,
        options.earthradius,
        options.schemeorder,
        options.nbsize,
        options.lttimes,
        options.lfrechet,
        options.tfieldsource + 1,
        options.lpaths,
        options.lcartesian,
        0,  # quiet=False (show warnings)
    )

    gdx, gdz, asgr, sgdl, sgs, earth, fom, snb, fsrt, cfd, wttf, wrgf, cart, quiet = (
        fmm.get_solver_options()
    )

    assert gdx == options.dicex
    assert gdz == options.dicey
    assert asgr == options.sourcegridrefine
    assert sgdl == options.sourcedicelevel
    assert sgs == options.sourcegridsize
    assert earth == options.earthradius
    assert fom == options.schemeorder
    assert snb == options.nbsize
    assert fsrt == options.lttimes
    assert cfd == options.lfrechet
    assert wttf == options.tfieldsource + 1
    assert wrgf == options.lpaths
    assert cart == options.lcartesian


def test_fmmin2d():
    os.chdir(Path(__file__).parent)
    fmm.fmmin2d()
    os.chdir(Path(__file__).parent.parent)


def test_read_solver_options():
    fmm.read_solver_options(FMINFILE)
    gdx, gdz, asgr, sgdl, sgs, earth, fom, snb, fsrt, cfd, wttf, wrgf, cart, quiet = (
        fmm.get_solver_options()
    )

    # read file manually and compare
    with open(FMINFILE, "r") as f:
        lines = f.readlines()

    assert gdx == int(lines[7].split()[0])
    assert gdz == int(lines[7].split()[1])

    assert asgr == int(lines[8].split()[0])

    assert sgdl == int(lines[9].split()[0])
    assert sgs == int(lines[9].split()[1])

    assert earth == float(lines[10].split()[0])

    assert fom == int(lines[11].split()[0])

    assert snb == float(lines[12].split()[0])

    assert fsrt == int(lines[16].split()[0])

    assert cfd == int(lines[18].split()[0])

    assert wttf == int(lines[20].split()[0])

    assert wrgf == int(lines[22].split()[0])


def test_read_velocity_model():
    fmm.read_velocity_model(VELMODELFILE)
    nvx, nvz, goxd, gozd, dvxd, dvzd, velv = fmm.get_velocity_model()

    # read file manually and compare
    with open(VELMODELFILE, "r") as f:
        lines = f.readlines()

    assert nvx == int(lines[0].split()[0])
    assert nvz == int(lines[0].split()[1])

    assert goxd == pytest.approx(float(lines[1].split()[0]))
    assert gozd == pytest.approx(float(lines[1].split()[1]))

    assert dvxd == pytest.approx(float(lines[2].split()[0]))
    assert dvzd == pytest.approx(float(lines[2].split()[1]))

    # The true model file has 2 columns where the velocity is stored
    # However, the Fortran code only reads the first column
    # DO i = 0, nvz + 1
    #     DO j = 0, nvx + 1
    #         READ (10, *) velv(i, j)
    #     END DO
    # END DO
    assert np.array_equal(
        velv.flatten(), np.loadtxt(VELMODELFILE, skiprows=3, usecols=0)
    )


def test_set_velocity_model():
    # fmm.set_velocity_model takes the cushionned velocity array
    # i.e. it expects something of shape (nvy + 2, nvx + 2)
    nvy, nvx = 10, 10
    extent = [1.0, 2.0, 3.0, 4.0]
    dlat, dlong = 0.1, 0.1
    vc = np.full((nvy + 2, nvx + 2), 2000.0)

    fmm.set_velocity_model(nvy, nvx, extent[3], extent[0], dlat, dlong, vc)
    nx, nz, goxd, gozd, dvxd, dvzd, velv = fmm.get_velocity_model()

    assert nx == nvy
    assert nz == nvx
    assert goxd == pytest.approx(extent[3])
    assert gozd == pytest.approx(extent[0])
    assert dvxd == pytest.approx(dlong)
    assert dvzd == pytest.approx(dlat)

    assert np.allclose(velv, vc, atol=1e-7)
    assert velv.shape == (nvy + 2, nvx + 2)


def test_read_sources():
    fmm.read_sources(SOURCESFILE)
    scx, scz = fmm.get_sources()
    sources = np.array([scx, scz]).T

    assert np.array_equal(sources, np.loadtxt(SOURCESFILE, skiprows=1))


def test_set_sources():
    sources = np.array([[0.1, 0.15], [0.2, 0.25]])
    scx = sources[:, 0]
    scy = sources[:, 1]
    fmm.set_sources(scy, scx)  # note the order of scx and scy
    _scx, _scz = fmm.get_sources()
    assert np.allclose(_scx, scy, atol=1e-7)
    assert np.allclose(_scz, scx, atol=1e-7)


def test_read_receivers():
    fmm.read_receivers(RECEIVERSFILE)
    rcx, rcz = fmm.get_receivers()
    receviers = np.array([rcx, rcz]).T

    assert np.array_equal(receviers, np.loadtxt(RECEIVERSFILE, skiprows=1))


def test_set_receivers():
    receivers = np.array([[0.8, 1], [1.0, 0.6]])
    rcx = receivers[:, 0]
    rcy = receivers[:, 1]
    fmm.set_receivers(rcy, rcx)  # note the order of rcx and rcy
    _rcx, _rcz = fmm.get_receivers()
    assert np.allclose(_rcx, rcy, atol=1e-7)
    assert np.allclose(_rcz, rcx, atol=1e-7)


def test_read_source_receiver_associations():
    fmm.read_sources(SOURCESFILE)
    fmm.read_receivers(RECEIVERSFILE)

    fmm.read_source_receiver_associations(ASSOCIATIONSFILE)
    srs = fmm.get_source_receiver_associations()

    # read file manually and compare
    with open(SOURCESFILE, "r") as f:
        lines = f.readlines()
        nsrc = int(lines[0].split()[0])
    with open(RECEIVERSFILE, "r") as f:
        lines = f.readlines()
        nrec = int(lines[0].split()[0])

    # have to manipulate the array to match the Fortran array
    associations = np.loadtxt(ASSOCIATIONSFILE, dtype=int).reshape(nsrc, nrec).T

    assert srs.shape == (nrec, nsrc)
    assert np.array_equal(srs, associations)


def test_track():
    # track is the main function
    #
    # get_raypaths, get_traveltimes, get_traveltime_fields, get_frechet_derivatives
    # only return values after track has been called
    options = FMMOptions()

    fmm.set_solver_options(
        np.int32(options.dicex),
        np.int32(options.dicey),
        np.int32(options.sourcegridrefine),
        np.int32(options.sourcedicelevel),
        np.int32(options.sourcegridsize),
        np.float32(options.earthradius),
        np.int32(options.schemeorder),
        np.float32(options.nbsize),
        np.int32(options.lttimes),
        np.int32(options.lfrechet),
        np.int32(options.tfieldsource + 1),
        np.int32(options.lpaths),
        np.int32(options.lcartesian),
        0,  # quiet=False (show warnings)
    )

    fmm.read_velocity_model(VELMODELFILE)
    fmm.read_sources(SOURCESFILE)
    fmm.read_receivers(RECEIVERSFILE)
    fmm.read_source_receiver_associations(ASSOCIATIONSFILE)

    fmm.allocate_result_arrays()

    fmm.track()

    ttimes = fmm.get_traveltimes()
    assert ttimes is not None

    paths = fmm.get_raypaths()
    assert paths is not None

    tfields = fmm.get_traveltime_fields()  ## failing silently here
    assert tfields is not None

    frechet = fmm.get_frechet_derivatives()
    assert frechet is not None

    fmm.deallocate_result_arrays()
